@@ -4,6 +4,7 @@
 from configparser import ConfigParser
 from logging import info, error
 import re
+import sys
 
 # Cloud cost allocation imports
 from cloud_cost_allocation.cost_items import CloudCostItem, ConsumerCostItem, CostItemFactory, \
@@ -29,15 +30,19 @@ class CloudCostAllocator(object):
 
     def allocate(self, consumer_cost_items: list[ConsumerCostItem], cloud_cost_items: list[CloudCostItem]) -> bool:
 
-        # Create the list of all cost items and process cloud tag selectors
-        info("Processing cloud tag selectors")
+        # Initiate cost items with cloud cost items
         cost_items = []
         cost_items.extend(cloud_cost_items)
+
+        # Process cloud tag selectors
+        info("Processing cloud tag selectors")
+        cloud_tag_selector_consumer_cost_items = []
         for consumer_cost_item in consumer_cost_items:
             if consumer_cost_item.provider_cost_allocation_type == "CloudTagSelector":
-                self.process_cloud_tag_selector(cost_items, consumer_cost_item, cloud_cost_items)
+                cloud_tag_selector_consumer_cost_items.append(consumer_cost_item)
             else:
                 cost_items.append(consumer_cost_item)
+        self.process_cloud_tag_selectors(cost_items, cloud_tag_selector_consumer_cost_items, cloud_cost_items)
 
         # Create and add cloud consumer cost item from tags
         info("Creating consumer cost items from tags")
@@ -184,8 +189,12 @@ class CloudCostAllocator(object):
                         actual_consumer_service_tag_key = consumer_service_tag_key
                         consumer_service = cost_item.tags[consumer_service_tag_key].strip().lower()
                         break
-                if consumer_service == "-":  # Ignore consumer service named "-"
-                    consumer_service = ""
+
+                # Check if consumer service must be ignored
+                if 'ConsumerServiceIgnoredValue' in self.config['TagKey']:
+                    for consumer_service_ignored_value in self.config['TagKey']['ConsumerServiceIgnoredValue'].split(","):
+                        if consumer_service == consumer_service_ignored_value.strip().lower():
+                            consumer_service = ""
 
                 # Check consumer instance
                 consumer_instance = ""
@@ -200,6 +209,7 @@ class CloudCostAllocator(object):
 
                 # Check product
                 # TODO: product dimension
+                # TODO: product ignored value
                 product = ""
                 actual_product_tag_key = ""
                 if 'Product' in self.config['TagKey']:
@@ -277,56 +287,126 @@ class CloudCostAllocator(object):
             self.service_instances[service_instance_id] = service_instance
         return self.service_instances[service_instance_id]
 
-    def process_cloud_tag_selector(self,
-                                   cost_items: list[CostItem],
-                                   consumer_cost_item: ConsumerCostItem,
-                                   cloud_cost_items: list[CloudCostItem]) -> None:
+    def process_cloud_tag_selectors(self,
+                                    cost_items: list[CostItem],
+                                    cloud_tag_selector_consumer_cost_items: list[ConsumerCostItem],
+                                    cloud_cost_items: list[CloudCostItem]) -> None:
 
-        # Process cost items
-        new_consumer_cost_items = {}  # Key is consumer service instance id
+        # Build evaluation error dict, used to avoid error streaming in case of incorrect cloud tag selector expression
+        # Key = provider service, provider instance, cloud tag selector expression, exception
+        # Value = count of errors
+        evaluation_error_dict = {}
+
+        # Build cloud tag dict: key is cloud cost item tag string; value is list of cloud cost items
+        # Since a lot of cloud cost items have the same tags, the goal is to minimize the number of tag evaluation
+        # contexts
+        cloud_tag_dict = {}
         for cloud_cost_item in cloud_cost_items:
+            tag_string = str(cloud_cost_item.tags)
+            if tag_string in cloud_tag_dict:
+                cloud_cost_item_list = cloud_tag_dict[tag_string]
+            else:
+                cloud_cost_item_list = []
+                cloud_tag_dict[tag_string] = cloud_cost_item_list
+            cloud_cost_item_list.append(cloud_cost_item)
 
-            # Skip if same service: cost allocation to own service is unwanted
-            # Check if cloud selector match
-            consumer_service_instance_id = ServiceInstance.get_id(cloud_cost_item.service, cloud_cost_item.instance)
-            cloud_tag_selector = consumer_cost_item.provider_cost_allocation_cloud_tag_selector
-            if cloud_cost_item.service != consumer_cost_item.provider_service and \
-                    cloud_cost_item.matches_cloud_tag_selector(cloud_tag_selector, consumer_cost_item.provider_service):
+        # Process cloud tag dict
+        for cloud_cost_item_list in cloud_tag_dict.values():
 
-                # Check if matching service instance was already processed
-                if consumer_service_instance_id in new_consumer_cost_items:
+            # Build variables from tags
+            eval_globals_dict = {}
+            for key, value in cloud_cost_item_list[0].tags.items():
+                eval_globals_dict[re.sub(r'[^a-z0-9_]', '_', key)] = value
 
-                    # Increase existing cost allocation key with the amortized cost of this cloud cost item
-                    new_consumer_cost_item = new_consumer_cost_items[consumer_service_instance_id]
-                    new_consumer_cost_item.provider_cost_allocation_key += cloud_cost_item.cloud_amortized_cost
+            # Process consumer cost items
+            for cloud_tag_selector_consumer_cost_item in cloud_tag_selector_consumer_cost_items:
 
-                else:
-                    # Create new consumer cost item for this instance
-                    # TODO: A partial clone like this should be done in the class itself
-                    #       else it becomes very hard to identify when new attributes are added that may need to be cloned as well
-                    #       Also, this breaks when subclassing with relevant new attributes
-                    #      -> Move to ConsumerCostItem class
-                    new_consumer_cost_item = self.cost_item_factory.create_consumer_cost_item()
-                    new_consumer_cost_item.date_str = cloud_cost_item.date_str
-                    new_consumer_cost_item.service = cloud_cost_item.service
-                    new_consumer_cost_item.instance = cloud_cost_item.instance
-                    new_consumer_cost_item.tags = consumer_cost_item.tags.copy()
-                    new_consumer_cost_item.provider_service = consumer_cost_item.provider_service
-                    new_consumer_cost_item.provider_instance = consumer_cost_item.provider_instance
-                    # TODO: dimension tags are ignored; should we create different consumer cost items for
-                    #  different dimensions?
-                    # TODO: split and dispatch the provider meter values based on amortized costs
-                    new_consumer_cost_item.provider_meters = consumer_cost_item.provider_meters.copy()
-                    new_consumer_cost_item.provider_cost_allocation_type = \
-                        consumer_cost_item.provider_cost_allocation_type
-                    new_consumer_cost_item.provider_tag_selector = consumer_cost_item.provider_tag_selector
-                    new_consumer_cost_item.provider_cost_allocation_key = cloud_cost_item.cloud_amortized_cost
-                    new_consumer_cost_item.provider_cost_allocation_cloud_tag_selector = \
-                        consumer_cost_item.provider_cost_allocation_cloud_tag_selector
+                # New consumer cost items created for this tag context and for this cloud tag selector
+                # There is a unique consumer cost item for a given consumer service instance
+                # TODO: Create different consumer cost items for different dimensions
+                new_consumer_cost_items = {}  # Key is (consumer) service instance id
 
-                    # Add new consumer cost item
-                    new_consumer_cost_items[consumer_service_instance_id] = new_consumer_cost_item
-                    cost_items.append(new_consumer_cost_item)
+                # Check if cloud selector match
+                match = False
+                try:
+                    # Eval is dangerous. Considerations:
+                    # - Possibly forbid cloud tag selectors
+                    # - Possibly whitelist cloud tag selectors using pattern matching
+                    match = eval(cloud_tag_selector_consumer_cost_item.provider_cost_allocation_cloud_tag_selector,
+                                 eval_globals_dict, {})
+                except:
+                    exception = sys.exc_info()[0]
+                    error_key =\
+                        cloud_tag_selector_consumer_cost_item.provider_service + chr(10) +\
+                        cloud_tag_selector_consumer_cost_item.provider_instance + chr(10) +\
+                        cloud_tag_selector_consumer_cost_item.provider_cost_allocation_cloud_tag_selector + chr(10) +\
+                        str(exception)
+                    if error_key in evaluation_error_dict:
+                        error_count = evaluation_error_dict[error_key]
+                    else:
+                        error_count = 0
+                    evaluation_error_dict[error_key] = error_count + 1
+                if match:
+
+                    # Process cloud cost items
+                    for cloud_cost_item in cloud_cost_item_list:
+
+                        # Skip if same service: cost allocation to own service is unwanted
+                        if cloud_cost_item.service != cloud_tag_selector_consumer_cost_item.provider_service:
+
+                            # Check if matching service instance was already processed
+                            consumer_cost_item_id =\
+                                ServiceInstance.get_id(cloud_cost_item.service, cloud_cost_item.instance)
+                            if consumer_cost_item_id in new_consumer_cost_items:
+
+                                # Increase existing cost allocation key with the amortized cost of this cloud cost item
+                                new_consumer_cost_item = new_consumer_cost_items[consumer_cost_item_id]
+                                new_consumer_cost_item.provider_cost_allocation_key +=\
+                                    cloud_cost_item.cloud_amortized_cost
+
+                            else:
+                                # Create new consumer cost item for this instance
+                                # TODO: A partial clone like this should be done in the class itself
+                                #       else it becomes very hard to identify when new attributes are added
+                                #       that may need to be cloned as well
+                                #       Also, this breaks when subclassing with relevant new attributes
+                                #      -> Move to ConsumerCostItem class
+                                new_consumer_cost_item = self.cost_item_factory.create_consumer_cost_item()
+                                new_consumer_cost_item.date_str = cloud_cost_item.date_str
+                                new_consumer_cost_item.service = cloud_cost_item.service
+                                new_consumer_cost_item.instance = cloud_cost_item.instance
+                                new_consumer_cost_item.tags = cloud_tag_selector_consumer_cost_item.tags.copy()
+                                new_consumer_cost_item.provider_service =\
+                                    cloud_tag_selector_consumer_cost_item.provider_service
+                                new_consumer_cost_item.provider_instance =\
+                                    cloud_tag_selector_consumer_cost_item.provider_instance
+                                # TODO: split and dispatch the provider meter values
+                                new_consumer_cost_item.provider_meters =\
+                                    cloud_tag_selector_consumer_cost_item.provider_meters.copy()
+                                new_consumer_cost_item.provider_cost_allocation_type = \
+                                    cloud_tag_selector_consumer_cost_item.provider_cost_allocation_type
+                                new_consumer_cost_item.provider_tag_selector =\
+                                    cloud_tag_selector_consumer_cost_item.provider_tag_selector
+                                new_consumer_cost_item.provider_cost_allocation_key =\
+                                    cloud_cost_item.cloud_amortized_cost
+                                new_consumer_cost_item.provider_cost_allocation_cloud_tag_selector = \
+                                    cloud_tag_selector_consumer_cost_item.provider_cost_allocation_cloud_tag_selector
+
+                                # Add new consumer cost item
+                                new_consumer_cost_items[consumer_cost_item_id] = new_consumer_cost_item
+                                cost_items.append(new_consumer_cost_item)
+
+        # Log evaluation errors
+        for error_key, error_count in evaluation_error_dict.items():
+            error_fields = error_key.split(chr(10))
+            provider_service = error_fields[0]
+            provider_instance = error_fields[1]
+            cloud_tag_selector = error_fields[2]
+            exception = error_fields[3]
+            error("Caught exception '" + exception + "' " + str(error_count) + " times " +
+                  "when evaluating ProviderCostAllocationCloudTagSelector '" + cloud_tag_selector +
+                  "' of ProviderService '" + provider_service +
+                  "' and of ProviderInstance '" + provider_instance + "'")
 
     def visit_for_allocation_and_set_allocated_cost(self, is_amortized_cost: bool, is_for_products: bool) -> None:
 
